"""
Models for TiddlyWeb Entities.

Psuedo-model classes representing the important
entities/data classes in the TiddlyWeb system.

These intentionally are limited, making no effort
to handle their own persistence.
"""
